﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Input;
using Uno.Disposables;
using Windows.UI.Core;

namespace Uno.UI.Samples.UITests.Helpers
{
	[Windows.UI.Xaml.Data.Bindable]
	public class ViewModelBase : INotifyPropertyChanged, IDisposable
	{
		public CoreDispatcher Dispatcher { get; }

		public event PropertyChangedEventHandler PropertyChanged;

		protected readonly CompositeDisposable Disposables = new CompositeDisposable();
		protected readonly CancellationToken CT;

		public ViewModelBase(CoreDispatcher dispatcher)
		{
			Dispatcher = dispatcher;

			var cts = new CancellationTokenSource();
			CT = cts.Token;

			Disposables.Add(Disposable.Create(() => cts.Cancel()));
		}

		protected void RaisePropertyChanged([CallerMemberName] string propertyName = "")
		{
			if (Dispatcher.HasThreadAccess)
			{
				RaiseEvent();
			}
			else
			{
				var _ = Dispatcher.RunAsync(CoreDispatcherPriority.Normal, RaiseEvent);
			}

			void RaiseEvent()
			{
				PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
				PropertyChanged?.Invoke(this, new PropertyChangedEventArgs($"Item[{propertyName}]"));
			}
		}

		protected static Command CreateCommand<T>(Action<T> action)
		{
			return new Command(x => action((T)x));
		}

		protected static Command CreateCommand(Action action)
		{
			return new Command(_ => action());
		}

		public void Dispose()
		{
			Disposables.Dispose();
		}

		public object this[string propertyName]
		{
			// This is a hack to adapt samples to an old way to build viewmodels.
			// You don't need to use this for new samples.
			get
			{
				var type = GetType();
				var pi = type.GetProperty(propertyName,
					BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.Public);
				var value = pi?.GetValue(this);
				return value;
			}
			set
			{
				var type = GetType();
				var pi = type.GetProperty(propertyName,
					BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.Public);
				pi?.SetValue(this, value);
			}
		}

		[Windows.UI.Xaml.Data.Bindable]
		public class Command : ICommand
		{
			private readonly Action<object> _action;
			private readonly Func<object, Task> _actionTask;
			private readonly Func<object, bool> _canExecute;
			private bool _manualCanExecute = true;

			public bool ManualCanExecute
			{
				get => _manualCanExecute;
				set
				{
					_manualCanExecute = value;
					CanExecuteChanged?.Invoke(this, EventArgs.Empty);
				}
			}

			private object _isExecuting = null;
			private static object _isExecutingNull = new object(); // this represent a null parameter when something is executing

			public Command(Action<object> action, Func<object, bool> canExecute = null)
			{
				_action = action ?? throw new ArgumentNullException(nameof(action));
				_canExecute = canExecute;
			}

			public Command(Func<object, Task> actionTask, Func<object, bool> canExecute = null)
			{
				_actionTask = actionTask ?? throw new ArgumentNullException(nameof(actionTask));
				_canExecute = canExecute;
			}

			public bool CanExecute(object parameter)
			{
				var canExecuteParameter = parameter ?? _isExecutingNull;

				return (_isExecuting != canExecuteParameter)
					&& (_canExecute?.Invoke(parameter) ?? true)
					&& _manualCanExecute;
			}

			public async void Execute(object parameter)
			{
				var isExecutingParameter = parameter ?? _isExecutingNull;

				if (_isExecuting == isExecutingParameter)
				{
					// This parameter is executing
					return;
				}

				try
				{
					_isExecuting = isExecutingParameter;
					CanExecuteChanged?.Invoke(this, EventArgs.Empty);
					if (_action != null)
					{
						_action(parameter);
					}
					else
					{
						await _actionTask(parameter);
					}
				}
				finally
				{
					_isExecuting = null;
					CanExecuteChanged?.Invoke(this, EventArgs.Empty);
				}
			}

			public event EventHandler CanExecuteChanged;
		}
	}
}
